from amino import Path, __, Just, List, L, Maybe, _

from ribosome.record import list_field, field, maybe_field, bool_field

from myo.record import Record, Named
from myo.ui.tmux.util import ident_field, Ident


default_signals = List('int', 'term', 'kill')


class Command(Named):
    line = field(str)
    log_path = maybe_field(Path)
    parser = maybe_field(str)
    transient = bool_field()
    langs = list_field()
    kill = bool_field()
    signals = list_field(str, initial=default_signals)

    @property
    def desc(self):
        info = self._info.mk_string(' ∘ ')
        tail = ': {}'.format(info) if info else ''
        return '{} {}{}'.format(self._type_desc, self.name, tail)

    @property
    def _info_fields(self):
        return List('langs', 'parser', 'log_path')

    @property
    def _info(self) -> List[str]:
        return self._info_fields // self._field_info

    def _field_info(self, name: str):
        v = getattr(self, name)
        fmt = '⎡{} -> {}⎤'.format
        if isinstance(v, Maybe):
            return v / L(fmt)(name, _)
        elif isinstance(v, List):
            return v.empty.no.maybe(fmt(name, v.mk_string(',')))
        else:
            return Just(fmt(name, v))

    @property
    def _type_desc(self):
        return 'C'


class VimCommand(Command):

    @property
    def _type_desc(self):
        return 'VC'


class ShellCommand(Command):
    shell = ident_field()

    @property
    def _type_desc(self):
        return 'SC'

    @property
    def _info_fields(self):
        return super()._info_fields + List('shell')


class Shell(ShellCommand):
    target = ident_field()

    @property
    def _type_desc(self):
        return 'S'


class Commands(Record):
    commands = list_field()
    history = list_field()
    transient_cmd = maybe_field(Command)

    def __add__(self, command: Command):
        return self.append1.commands(command)

    def __getitem__(self, ident: Ident):
        return (
            self.commands.find(__.has_ident(ident))
            .to_either('no command with ident {}'.format(ident))
        )

    def __str__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ','.join(map(repr, self.commands))
        )

    def shell(self, ident):
        pred = lambda a: isinstance(a, Shell) and a.has_ident(ident)
        return self.commands.find(pred)

    def add_history(self, uuid):
        return self.append1.history(uuid)

    def transient_command(self, cmd):
        return self.set(transient_cmd=Just(cmd))

    @property
    def latest_command(self):
        def fetch(ident):
            return (
                self.transient_cmd.filter(__.has_ident(ident))
                .or_else(lambda: self[ident])
            )
        return self.history.head // fetch

__all__ = ('Commands', 'Command')



from ribosome import Machine, NvimFacade, RootMachine
from ribosome.nvim import HasNvim

from ribosome.machine.state import SubMachine, SubTransitions

from myo.logging import Logging
from myo.env import Env


class MyoComponent(SubMachine, HasNvim, Logging):

    def __init__(self, vim: NvimFacade, parent=None, title=None) -> None:
        Machine.__init__(self, parent, title=title)
        HasNvim.__init__(self, vim)


class MyoState(RootMachine, Logging):
    _data_type = Env

    @property
    def title(self):
        return 'myo'


class MyoTransitions(SubTransitions, HasNvim, Logging):

    def __init__(self, machine, *a, **kw):
        SubTransitions.__init__(self, machine, *a, **kw)
        HasNvim.__init__(self, machine.vim)

__all__ = ('MyoComponent', 'MyoState', 'MyoTransitions')



import abc
from pathlib import Path
import tempfile
import os

from psutil import Process

from libtmux.pane import Pane as LTPane

from amino.task import task

from amino import __, List, Boolean, Maybe, _, Map, Just, Try, L
from amino.lazy import lazy

from ribosome.record import maybe_field, either_field, bool_field

from myo.ui.tmux.view import View
from myo.ui.tmux.adapter import Adapter
from myo.util import parse_int
from myo.ui.tmux.util import parse_window_id, parse_pane_id, parse_session_id


class Pane(View):
    id = maybe_field(int)
    pid = maybe_field(int)
    shell_pid = maybe_field(int)
    window_id = either_field(int)
    session_id = either_field(int)
    log_path = maybe_field(Path)
    pin = bool_field()
    focus = bool_field()
    kill = bool_field()

    @property
    def id_s(self):
        return self.id / '%{}'.format

    @property
    def desc(self):
        id = self.id / ' %{}'.format | ''
        pid = self.pid / ' | pid -> {}'.format | ''
        pin = ' !' if self.pin else ''
        return 'P{} \'{}\'{}{} {}'.format(id, self.name, pin, pid,
                                          self.size_desc)

    def open(self, pa: 'PaneAdapter'):
        return self.set(id=pa.id_i.to_maybe, shell_pid=pa.pid.to_maybe,
                        session_id=pa.session_id_i, window_id=pa.window_id_i)

    @property
    def _str_extra(self):
        return self.id.to_list + super()._str_extra


class VimPane(Pane):
    pane_name = '<vim>'

    def __new__(cls, *a, name=None, **kw):
        n = Maybe(name) | VimPane.pane_name
        return super().__new__(cls, *a, name=n, **kw)

    @property
    def desc(self):
        return 'V{}'.format(super().desc)


class PaneI(metaclass=abc.ABCMeta):

    @abc.abstractproperty
    def _raw_pid(self) -> Maybe[int]:
        ...

    @lazy
    def pid(self):
        return self._raw_pid // parse_int

    @property
    def command_pid(self) -> Maybe[int]:
        return (
            self.pid //
            L(Try)(Process, _) /
            __.children() /
            List.wrap //
            _.head /
            _.pid //
            Maybe
        )


class PaneData(PaneI):

    def __init__(self, data: Map) -> None:
        self.data = data

    def attr(self, name):
        return (
            self.data.get(name)
            .or_else(self.data.get('pane_{}'.format(name)))
        )

    @lazy
    def id(self):
        return self.attr('id')

    @property
    def id_i(self):
        return self.id // parse_pane_id

    @lazy
    def window_id(self):
        return self.attr('window_id')

    @property
    def window_id_i(self):
        return self.window_id // parse_window_id

    @lazy
    def session_id(self):
        return self.attr('session_id')

    @property
    def session_id_i(self):
        return self.session_id // parse_session_id

    @property
    def _raw_pid(self):
        return self.data.get('pane_pid')


class NativePane(LTPane):

    def __init__(self, window=None, **kwargs):
        if not window:
            raise ValueError('Pane must have ``Window`` object')
        self.window = window
        self.session = self.window.session
        self.server = self.session.server
        self._pane_id = kwargs['pane_id']
        self._data = Map(kwargs)

    @property
    def _info(self):
        return self._data


class PaneAdapter(Adapter, PaneI):

    @lazy
    def id(self):
        return self.native._pane_id

    @lazy
    def id_i(self):
        return parse_pane_id(self.id)

    @property
    def _raw_pid(self):
        return Try(lambda: self.native.pid)

    @task
    def resize(self, size, horizontal):
        f = __.set_width if horizontal else __.set_height
        return f(size)(self.native)

    @property  # type: ignore
    @task
    def kill(self):
        return self.cmd('kill-pane')

    def split(self, horizontal):
        return self.native.split_window(vertical=not horizontal)

    @lazy
    def width(self):
        return int(self.native.width)

    @lazy
    def height(self):
        return int(self.native.height)

    @lazy
    def size(self):
        return self.width, self.height

    @lazy
    def left(self):
        return int(self.native.left)

    @lazy
    def top(self):
        return int(self.native.top)

    @lazy
    def position(self):
        return self.left, self.top

    def send_keys(self, cmd, enter=True, suppress_history=True):
        return self.native.send_keys(cmd, enter=enter,
                                     suppress_history=suppress_history)

    @property
    def capture(self):
        return List.wrap(self.cmd('capture-pane', '-p'))

    @property
    def session_id(self):
        return self.native.session.id

    @lazy
    def session_id_i(self):
        return parse_session_id(self.session_id)

    @lazy
    def window_id(self):
        return self.native.window.id

    @lazy
    def window_id_i(self):
        return parse_window_id(self.window_id)

    def __repr__(self):
        return 'PA({}, {}, {})'.format(self.id, self.size, self.position)

    @property
    def running(self) -> Boolean:
        return self.command_pid.is_just

    @property
    def not_running(self) -> Boolean:
        return Boolean(not self.running)

    @property
    def _pipe_filter(self):
        return 'sed -u -e \'s/\r//g\' -e \'s/\x1b\[[0-9;?]*[mlK]//g\''

    def pipe(self, base):
        uid = os.getuid()
        tmpdir = Path(tempfile.gettempdir()) / 'myo-{}'.format(uid) / base
        tmpdir.mkdir(exist_ok=True, parents=True)
        (fh, fname) = tempfile.mkstemp(prefix='pane-', dir=str(tmpdir))
        self.cmd('pipe-pane', '{} > {}'.format(self._pipe_filter, fname))
        return Just(Path(fname))

    def move_above(self, other: 'PaneAdapter'):
        self.cmd('move-pane', '-b', '-s', other.id)

    def move_below(self, other: 'PaneAdapter'):
        self.cmd('move-pane', '-s', other.id)

    def swap(self, other: 'PaneAdapter'):
        self.cmd('swap-pane', '-d', '-s', other.id)

    def __eq__(self, other):
        return isinstance(other, PaneAdapter) and self.id == other.id

    @property
    def active(self):
        return self.native.active == '1'

    def focus(self):
        self.native.select_pane()

    def run_command(self, line):
        self.quit_copy_mode()
        self.send_keys(line, suppress_history=False)

    @property
    def in_copy_mode(self):
        return self.native.in_mode == '1'

    def quit_copy_mode(self):
        if self.in_copy_mode:
            self.send_keys('C-c')

__all__ = ('Pane', 'VimPane', 'PaneAdapter')



from integration._support.base import MyoIntegrationSpec

from amino.test.path import fixture_path
from amino import Map, Just, List, _
from amino.test import temp_file, later
from amino.lazy import lazy

from ribosome.test.integration import main_looped
from ribosome.machine.scratch import Mapping

from myo.command import Command
from myo.plugins.core.message import ParseOutput
from myo.output.machine import error_filtered_result_empty
from myo.plugins.core.main import error_no_output_events

_trace_len = 3
_line_count = _trace_len * 4 + 5


class ParseHelpers:

    def _modifiable(self, val):
        later(lambda: self.vim.buffer.options('modifiable').should.contain(val)
              )

    @lazy
    def _file(self):
        path = temp_file('core', 'parse', 'python', 'code.py')
        path.write_text(self._python_code.join_lines)
        return path

    @property
    def _python_func_name(self):
        return 'func_name'

    @lazy
    def _python_code(self):
        return List(
            'import functools',
            'def {}(a, b=1):'.format(self._python_func_name),
            '    return 1'
        )

    def _frame(self, i):
        r = lambda: List.random_string(4)
        tmpl = '  File "{}", line 2, in {}'
        return List(
            tmpl.format(self._file, self._python_func_name),
            '    {}()'.format(r())
        )

    def _mk_trace_with(self, length):
        t = List.range(length) // self._frame
        err = 'AttributeError: butt'
        return t.cat(err)

    @property
    def _mk_trace(self):
        return self._mk_trace_with(_trace_len)


class BasicParseSpec(MyoIntegrationSpec, ParseHelpers):

    @main_looped
    def native(self):
        cmd = Command(name='a', line='a', parser=Just('compiler'))
        fix = fixture_path('parse', 'err1')
        output = List.wrap(fix.read_text().splitlines())
        errorformat = List(
            '%AIn %f:%l',
            '%C  %.%#',
            '%+Z%.%#Error: %.%#'
        ).mk_string(',')
        self.vim.buffer.options.set('errorformat', errorformat)
        msg = ParseOutput(cmd, output, fix, Map())
        self.plugin.myo_start()
        self.root.send(msg)
        self._await()
        qf = self.vim.call('getqflist') | []
        qf.should_not.be.empty

    @main_looped
    def twice(self):
        self.vim.vars.set_p('jump_to_error', False)
        cmd = Command(name='a', line='a', langs=List('python'))
        msg = ParseOutput(cmd, self._mk_trace, None, Map())
        self.plugin.myo_start()
        self.root.send_sync(msg)
        output_machine = self.root.sub[-1]
        self._modifiable(False)
        self.root.send_sync(Mapping(output_machine.uuid, 'q'))
        self._modifiable(True)
        self.root.send_sync(msg)
        self._modifiable(False)

    @main_looped
    def empty(self):
        cmd = Command(name='a', line='a', langs=List('python'))
        msg = ParseOutput(cmd, List(), None, Map())
        self.plugin.myo_start()
        self.root.send_sync(msg)
        later(lambda:
              self.vim.messages.should.contain(error_no_output_events))
        self.vim.windows.should.have.length_of(1)


class ParseSpecBase(MyoIntegrationSpec, ParseHelpers):

    def _pre_start(self):
        super()._pre_start()
        self.vim.vars.set_p('jump_to_error', False)

    def _cursor(self, x, y):
        later(lambda: self.vim.window.cursor.should.equal(List(x, y)))


class PythonParseSpec(ParseSpecBase):

    def _pre_start(self):
        super()._pre_start()
        self.vim.vars.set_p('output_reifier',
                            'py:myo.output.reifier.base.LiteralReifier')

    def _parse(self, output, parse_opt=Map()):
        cmd = Command(name='a', line='a', langs=List('python'))
        msg = ParseOutput(cmd, output, None, parse_opt)
        self.root.send_sync(msg)
        return output

    def _run(self, output, parse_opt=Map()):
        self.plugin.myo_start()
        return self._parse(output, parse_opt)

    def _init(self, parse_opt=Map()):
        trace1 = self._mk_trace
        trace2 = self._mk_trace
        output = trace1 + trace2
        return self._run(output, parse_opt)

    def _go(self, line_count=_line_count, parse_opt=Map()):
        output = self._init(parse_opt)
        self.vim.buffer.content.should.contain(output[-1])
        self.vim.buffer.options('modifiable').should.contain(False)
        self.vim.buffer.content.should.have.length_of(line_count)

    def _check_jumped(self):
        self.vim.window.buffer.name.should.equal(str(self._file))
        self._cursor(2, 0)

    def jump(self):
        self._go()
        output_machine = self.root.sub[-1]
        self.vim.window.set_cursor(4)
        self.root.send_sync(Mapping(output_machine.uuid, '%cr%'))
        self.vim.windows.should.have.length_of(2)
        self._check_jumped()

    def filter(self):
        filters = List('py:integration.core.parse_spec._filter1')
        self._go(_trace_len * 2 + 3, Map(output_filters=filters))

    def initial_pos(self):
        self.vim.vars.set_p('first_error',
                            ['py:integration.core.parse_spec._first_error'])
        self._go()
        self.vim.window.line.should.contain(_line_count - 2)

    def initial_pos_default(self):
        self._go()
        self.vim.window.line.should.contain(_line_count - 2)

    def initial_pos_first(self):
        self.vim.vars.set_p('first_error', ['s:first'])
        self._go()
        self.vim.window.line.should.contain(3)

    def initial_pos_with_jump(self):
        self.vim.vars.set_p('jump_to_error', True)
        self.vim.vars.set_p('first_error',
                            ['py:integration.core.parse_spec._first_error'])
        self._init()
        self._check_jumped()

    def default_jump(self):
        self.vim.vars.set_p('jump_to_error', True)
        self._init()
        self._check_jumped()

    def lang_reifier(self):
        self.vim.vars.set_p('output_reifier',
                            'py:myo.output.reifier.python.Reifier')
        trace1 = self._mk_trace_with(1)
        self._run(trace1)
        self._cursor(1, 1)

    def twice(self):
        t1 = self._mk_trace_with(1)
        t2 = self._mk_trace_with(1)
        def check(t):
            later(lambda: self.vim.buffer.content.lift(-2)
                  .should.equal(t.lift(-2)))
        self._run(t1)
        check(t1)
        self._parse(t2)
        check(t2)

    def _syntax(self):
        from amino import log
        self.vim.vars.set_p('output_reifier', '')
        self.vim.vars.set_p('jump_to_error', True)
        self.vim.cmd_sync('hi Error cterm=bold ctermfg=1')
        self._init()
        self._wait(3)
        log.info(self.vim.cmd_output('syntax').join_lines)
        self._check_jumped()

    def duplicate(self):
        t = self._mk_trace_with(2)
        trace = t + t
        self._run(trace)
        later(lambda: self.vim.buffer.content[2:].should.equal(t))

    def _empty_filtered(self):
        filters = List('py:integration.core.parse_spec._filter_empty')
        self.vim.vars.set_p('output_reifier',
                            'py:myo.output.reifier.python.Reifier')
        self._init(Map(output_filters=filters))
        later(lambda:
              self.vim.messages.should.contain(error_filtered_result_empty))


def _filter1(result):
    e = result.events
    e2 = e[:1] + e[1 + _trace_len * 2 + 2:]
    return result.set(events=e2)


def _filter_empty(result):
    return result.set(events=List())


def _first_error(a):
    return Just((_line_count - 3, 1))


class SbtParseSpec(ParseSpecBase):

    @lazy
    def _scala_file(self):
        path = temp_file('core', 'parse', 'sbt', 'code.scala')
        path.write_text(self._scala_code.join_lines)
        return path

    @property
    def _line(self):
        return 3

    @property
    def _scala_func_name(self):
        return 'scalafunc'

    @lazy
    def _scala_code(self):
        return List(
            'import foo.bar',
            'class Foo {',
            'def {}(a: Int): Int = a'.format(self._scala_func_name),
            '}',
        )

    @property
    def _mk_scala_error(self):
        l = 4
        r = lambda: List.random_string(l)
        tmpl = '[error] {}:{}: butt'
        return List(
            tmpl.format(self._scala_file, self._line),
            '[error] {}()'.format(r()),
            '[error] {}^'.format(' ' * l),
        )

    def _mk_scala_errors(self, count):
        return List.gen(count, lambda: self._mk_scala_error)

    def _run(self, output):
        cmd = Command(name='a', line='a', langs=List('sbt'))
        msg = ParseOutput(cmd, output, None, Map())
        self.plugin.myo_start()
        self.root.send_sync(msg)

    def complete(self):
        self.vim.vars.set_p('output_reifier',
                            'py:myo.output.reifier.base.LiteralReifier')
        output = self._mk_scala_errors(1)
        self._run(output)
        output_machine = self.root.sub[-1]
        self.vim.buffer.content.should.contain(output[-1])
        self.vim.buffer.options('modifiable').should.contain(False)
        self.vim.window.set_cursor(4)
        self.root.send_sync(Mapping(output_machine.uuid, '%cr%'))
        self.vim.windows.should.have.length_of(2)
        self.vim.window.buffer.name.should.equal(str(self._scala_file))
        self._cursor(3, 4)

    def filter_format(self):
        def check():
            c = self.vim.buffer.content
            c.should.have.length_of(8)
            c.head.should.contain('{}  3'.format(self._scala_file.name))
            self.vim.window.height.should.contain(8)
        filters = List('py:integration.core.parse_spec._filter')
        formatters = List('py:integration.core.parse_spec._format')
        trunc = 'py:integration.core.parse_spec._trunc'
        self.vim.vars.set_p('output_filters', filters)
        self.vim.vars.set_p('output_formatters', formatters)
        self.vim.vars.set_p('path_truncator', trunc)
        output = self._mk_scala_errors(4)
        self._run(output)
        later(check)

    def _syntax(self):
        from amino import log
        self.vim.cmd_sync('hi Error cterm=bold ctermfg=1')
        output = self._mk_scala_errors(1)
        self._run(output)
        self._wait(2)
        log.info(self.vim.cmd_output('syntax').join_lines)


def _filter(r):
    return r.modder.events(_[2:])


def _format(r):
    return r


def _trunc(path):
    return path.name

__all__ = ('BasicParseSpec', 'PythonParseSpec', 'SbtParseSpec')



import abc
from typing import Tuple, Callable

from amino import _, Path, List, __, Empty, Just, L
from amino.lazy import lazy

from ribosome.record import (field, list_field, dfield, any_field, RecordMeta,
                             maybe_field)

from myo.record import Record
from myo.util import parse_int


class OutputEntry(Record):
    text = field(str)

    @property
    def _str_name(self):
        return self._name.replace('Entry', '')

    def lines(self, event, group=Empty()):
        return self.format_lines(event, _.text, group)

    def format_lines(self, event, f: Callable, group=Empty()):
        return List(OutputLine.create(f(self), event | self, Just(self),
                                      group=group))


class OutputLine(Record):
    text = field(str)
    target = any_field()
    entry = maybe_field(OutputEntry)
    indent = dfield(0)
    group = maybe_field(str)

    def __str__(self):
        return '{}({})'.format(self._name, self.text)

    @staticmethod
    def create(text, target, entry=Empty(), group=Empty()):
        return OutputLine(text=text, target=target, entry=entry, group=group)

    @property
    def formatted(self):
        return '{}{}'.format(' ' * self.indent, self.text)

    @property
    def syntax_group(self):
        return self.group.o(self.entry / __._name.replace('Entry', ''))


class EmptyLine(OutputLine):

    @staticmethod
    def create(target):
        return EmptyLine(text='', target=target)


class ErrorEntry(OutputEntry):
    error = field(str)


class CodeEntry(OutputEntry):
    code = field(str)


def _parse_line(data):
    return data if isinstance(data, int) else parse_int(data) | 0


class LocationMeta(abc.ABCMeta, RecordMeta):
    pass


class Location:

    def file_path(self) -> Path:
        ...

    def coords(self) -> Tuple[int, int]:
        ...


class PositionEntry(OutputEntry, Location):
    path = field(Path, factory=Path)
    line = field(int, factory=_parse_line)
    col = dfield(0)

    @property
    def _str_extra(self):
        return List(self.path.name, self.coords)

    @property
    def file_path(self):
        return self.path

    @property
    def coords(self):
        return (self.line, self.col)


class OutputEvent(Record):
    head = list_field(str)
    entries = list_field(OutputEntry)

    @property
    def _str_extra(self):
        return self.entries

    @property
    def _target(self):
        return Just(self)

    @property
    def lines(self):
        return (
            (self.head / L(OutputLine.create)(_, self)) +
            (self.entries // __.lines(self._target))
        )


class MultiEvent(OutputEvent):

    @property
    def _target(self):
        return Empty()


class ParseResult(Record):
    head = list_field(str)
    events = list_field(OutputEvent)
    langs = list_field(str)

    @lazy
    def lines(self):
        return (
            (self.head / L(OutputLine.create)(_, self)) +
            (self.events // _.lines)
        )

    @property
    def display_lines(self):
        return self.lines / _.formatted

    def target_for_line(self, line):
        return self.lines.lift(line) / _.target

__all__ = ('OutputEntry', 'OutputEvent', 'ParseResult')



from collections import namedtuple

from ribosome.machine import may_handle, handle
from ribosome.machine.transition import Fatal
from ribosome.util.callback import parse_callback_spec

from myo.state import MyoComponent, MyoTransitions

from amino import L, _, List, Try, __, Maybe, curried, Map, Right, I
from myo.command import Command, VimCommand, ShellCommand, Shell
from myo.util import amend_options
from myo.plugins.core.message import Parse, ParseOutput, StageI
from myo.plugins.command.message import (Run, ShellRun, Dispatch, AddCommand,
                                         AddShellCommand, AddShell,
                                         AddVimCommand, SetShellTarget,
                                         CommandExecuted, RunTest, RunVimTest,
                                         CommandAdded, CommandShow, RunLatest,
                                         LoadHistory, StoreHistory, RunLine)
from myo.plugins.command.util import assemble_vim_test_line

RunInShell = namedtuple('RunInShell', ['shell'])


class CommandTransitions(MyoTransitions):
    _test_cmd_name = '<test>'

    def _add(self, tpe: type, **strict):
        return self._from_opt(tpe, **strict) / self._add_command

    def _add_command(self, cmd: Command):
        return List(
            self.data.cat(cmd), CommandAdded(cmd, options=self.msg.options)
        )

    @may_handle(StageI)
    def stage_i(self):
        line = self.vim.vars(self._last_test_line_var) | ''
        shell = self.vim.vars(self._last_test_shell_var)
        opt1 = Map(line=line)
        opt = shell / (lambda a: opt1.cat(('shell', a))) | opt1
        return (
            AddShellCommand(name=self._test_cmd_name, options=opt),
            LoadHistory()
        )

    @property
    def _history_var(self):
        return 'Myo_history'

    @property
    def _last_test_line_var(self):
        return 'Myo_last_test_line'

    @property
    def _last_test_shell_var(self):
        return 'Myo_last_test_shell'

    @handle(LoadHistory)
    def load_history(self):
        return (
            self.vim.vars.s(self._history_var) /
            __.split(',') /
            List.wrap /
            (lambda a: self.data.commands.set(history=a)) /
            self.data.setter.commands
        )

    @may_handle(StoreHistory)
    def store_history(self):
        names = (
            self.data.commands.history /
            self.data.command //
            _.to_maybe /
            _.name
        )
        self.vim.vars.set(self._history_var, names.mk_string(','))

    @handle(AddCommand)
    def add_command(self):
        return self._add(Command, name=self.msg.name)

    @may_handle(AddVimCommand)
    def add_vim_command(self):
        return self._add(VimCommand, name=self.msg.name)

    @handle(AddShellCommand)
    def add_shell_command(self):
        return self._add(ShellCommand, name=self.msg.name)

    @handle(AddShell)
    def add_shell(self):
        return self._from_opt(Shell, name=self.msg.name) / (
            lambda shell: (
                self._add_command(shell) +
                (shell.target / L(SetShellTarget)(shell, _) / _.pub).to_list
            )
        )

    @may_handle(CommandAdded)
    def command_added(self):
        if self.msg.options.get('start').exists(I):
            return Run(self.msg.command.uuid)

    @handle(Run)
    def run(self):
        return (
            self.data.command(self.msg.command) /
            L(Dispatch)(_, self.msg.options)
        )

    @handle(RunLine)
    def run_line(self):
        return self.msg.args.detach_head.map2(
            lambda target, line:
            Dispatch(ShellCommand(name='run_line', line=line.mk_string(' ')),
                     Map(target=target))
        )

    @handle(ShellRun)
    def run_in_shell(self):
        line = self.msg.options.get('line') | ''
        def send(cmd):
            return (
                self.data.transient_command(cmd),
                Dispatch(cmd, self.msg.options)
            )
        # double Maybe is necessary - the outer indicates success, the
        # inner is the type of the ShellCommand field
        return (
            self.data.shell(self.msg.shell) /
            _.uuid /
            Maybe /
            (lambda a: ShellCommand(name='shell_run', shell=a, line=line,
                                    transient=True)) /
            send
        )

    @handle(RunLatest)
    def run_latest(self):
        return (
            self.data.commands.latest_command /
            Dispatch
        )

    @handle(Dispatch)
    def dispatch(self):
        cmd = self.msg.command
        self.vim.vars.set_p('last_command', dict(name=cmd.name))
        self.vim.pautocmd('RunCommand')
        return (
            self.data.dispatch_message(cmd, self.msg.options) /
            _.pub /
            (lambda a: (a, CommandExecuted(cmd)))
        )

    @may_handle(CommandExecuted)
    def command_executed(self):
        new = self.data.commands.add_history(self.msg.command.uuid)
        return self.data.set(commands=new), StoreHistory()

    @handle(Parse)
    def parse(self):
        cmd = (
            self.msg.options.get('command')
            .cata(self.data.command, lambda: self.data.commands.latest_command)
            .to_either('invalid command name or empty history')
        )
        def find_log(c):
            return (
                (c.shell // self.data.shell | c)
                .log_path.to_either('{} has no log'.format(c))
            )
        log_path = cmd // find_log
        def parse(c, l):
            self.log.debug('parsing {} from {}'.format(c, l))
            return (
                Try(l.read_text) /
                __.splitlines() /
                List.wrap /
                L(ParseOutput)(c, _, l, self.msg.options) /
                _.pub
            )
        return (cmd & log_path).flat_map2(parse).lmap(Fatal)

    @handle(RunTest)
    def run_test(self):
        opt = self.msg.options
        return (
            opt.get('ctor').to_either(Fatal('no test ctor specified')) //
            self._assemble //
            self._run_test_line(opt - 'ctor')
        ).lmap(Fatal)

    @handle(RunVimTest)
    def run_vim_test(self):
        return (
            assemble_vim_test_line(self.vim) //
            __.head.to_either('vim-test failed') //
            self._run_test_line(self.msg.options)
        )

    @may_handle(CommandShow)
    def show(self):
        msg = self.data.commands.commands / _.desc
        self.log.info(msg.join_lines)

    def _assemble(self, ctor):
        return parse_callback_spec(ctor).join / (lambda a: a())

    @curried
    def _run_test_line(self, options, line):
        glangs = self.vim.buffer.vars.pl('test_langs')
        langs = options.get('langs').or_else(glangs) | List()
        shell = self.vim.buffer.pvar_or_global('test_shell')
        target = self.vim.buffer.pvar_or_global('test_target')
        opt = amend_options(options, 'shell', shell)
        opt2 = amend_options(opt, 'target', target)
        self.vim.vars.set(self._last_test_line_var, line)
        shell / L(self.vim.vars.set)('Myo_last_test_shell', _)
        def dispatch(data):
            return Right(data) & (data.command(self._test_cmd_name) /
                                  L(Dispatch)(_, opt2))
        return (
            self.data.command_lens(self._test_cmd_name)
            .to_either('no test command') /
            __.modify(__.set(line=line, langs=langs)) //
            dispatch
        )


class Plugin(MyoComponent):
    Transitions = CommandTransitions

    def new_state(self):
        pass

__all__ = ('Plugin', 'AddVimCommand', 'AddCommand', 'Run', 'AddShell',
           'AddShellCommand', 'SetShellTarget')



from pathlib import Path

from toolz.itertoolz import cons

from amino import List

from ribosome import NvimFacade

from myo.env import Env
from myo.state import MyoState


class Myo(MyoState):

    def __init__(
            self,
            vim: NvimFacade,
            config_path: Path=Path('/dev/null'),
            plugins: List[str]=List(),
    ) -> None:
        self._config_path = config_path
        core = 'myo.plugins.core'
        super().__init__(vim, List.wrap(cons(core, plugins)))

    @property
    def init(self):
        return Env(  # type: ignore
            config_path=self._config_path,
        )

__all__ = ('Myo',)



from typing import Callable

from myo.ui.tmux.view import View
from myo.ui.tmux.pane import Pane
from myo.ui.tmux.util import Ident

from lenses import lens

from amino import _, __, List, Just, Maybe
from amino.lazy import lazy

from ribosome.record import dfield, list_field, field, Record


class LayoutDirection(Record):
    id = field(str)


class LayoutDirections:
    vertical = LayoutDirection(id='vertical')
    horizontal = LayoutDirection(id='horizontal')
    all = List(vertical, horizontal)
    all_ids = all / _.id  # type: List[str]

    @staticmethod
    def parse(s: str):
        return (LayoutDirections.horizontal if s == 'horizontal' else
                LayoutDirections.vertical)


class Layout(View):
    panes = list_field()
    layouts = list_field()
    direction = dfield(LayoutDirections.vertical)

    def __repr__(self):
        return '{}({}, {!r}, {!r})'.format(self.__class__.__name__, self.name,
                                           self.panes, self.layouts)

    __str__ = __repr__

    def pane_index(self, f: Callable[[Pane], bool]):
        self.panes.index_where(f)

    def find_sub_pred(self, sub: Callable[['Layout'], List[View]],
                      pred: Callable[[View], bool]):
        return (sub(self).find(pred)
                .or_else(lambda: self._find_sub_in_layouts(sub, pred)))

    def _find_sub_in_layouts(self, sub: Callable[['Layout'], List[View]],
                             pred: Callable[[View], bool]):
        return self.layouts.find_map(__.find_sub_pred(sub, pred))

    def find_sub(self, sub: Callable[['Layout'], List[View]], ident: Ident):
        return self.find_sub_pred(sub, __.has_ident(ident))

    def find_layout(self, ident: Ident):
        return self.find_sub(_.layouts, ident)

    def find_pane(self, ident: Ident) -> Maybe[Pane]:
        return self.find_sub(_.panes, ident)

    def find_view(self, ident: Ident):
        return self.find_pane(ident).or_else(self.find_layout(ident))

    def view_lens(self, attr: Callable[['Layout'], List[View]],
                  f: Callable[[View], bool]):
        h = lambda l: l.layouts.find_lens(g) / lens().layouts.add_lens
        def g(l):
            return (
                attr(l).lens(f)
                .map(attr(lens()).add_lens)
                .or_else(lambda: h(l))
            )
        return g(self)

    def layout_lens(self, f: Callable[['Layout'], bool]):
        return Just(lens()) if f(self) else self.view_lens(_.layouts, f)

    def pane_lens(self, f: Callable[[Pane], bool]):
        return self.view_lens(_.panes, f)

    def replace_pane(self, pane: Pane):
        panes = self.panes.replace_where(pane)(__.has_ident(pane.ident))
        return self.set(panes=panes)

    @lazy
    def views(self):
        return self.panes + self.layouts

    @lazy
    def horizontal(self):
        return self.direction == LayoutDirections.horizontal

    @lazy
    def weights(self):
        w = self.views / _.weight
        total = sum(w.join)
        return w / (lambda a: a / (_ / total))

    @lazy
    def actual_min_sizes(self):
        return self.views / (lambda a:
                             a.effective_fixed_size.or_else(a.min_size) | 0)

    @lazy
    def actual_max_sizes(self):
        return self.views / (lambda a:
                             a.effective_fixed_size.or_else(a.max_size) | 999)

    @property
    def desc(self):
        dir = 'H' if self.horizontal else 'V'
        return 'L {} \'{}\' {}'.format(dir, self.name, self.size_desc)

    @property
    def all_panes(self):
        return self.panes + (self.layouts // _.all_panes)


class VimLayout(Layout):

    @property
    def desc(self):
        return 'V{}'.format(super().desc)

__all__ = ('Layout', 'LayoutDirection', 'VimLayout')



from networkx import DiGraph

from amino.lazy import lazy
from amino import List, _

from ribosome.record import field, maybe_field

from myo.output.data import (PositionEntry, OutputEntry, OutputEvent,
                             Location, CodeEntry as CodeEntryBase)
from myo.output.parser.base import EdgeData, SimpleParser


class FileEntry(PositionEntry):
    error = field(str)
    tag = field(str)

    @property
    def _str_extra(self):
        return super()._str_extra.cat(self.error)


class CodeEntry(CodeEntryBase):
    tag = field(str)


class ColEntry(OutputEntry):
    ws = field(str)
    tag = field(str)

    @property
    def _str_extra(self):
        return List(self.col)

    @property
    def col(self):
        return len(self.ws)


_msg_type = '\[(?P<tag>\w+)\] '
_file = EdgeData(
    r='^{}\s*(?P<path>[^:]+):(?P<line>\d+):\s*(?P<error>.*?);?$'
    .format(_msg_type),
    entry=FileEntry
)
_code = EdgeData(
    r='^{}(?P<code>\s*.+)'.format(_msg_type),
    entry=CodeEntry
)
_col = EdgeData(
    r='^{}(?P<ws>\s*)\^\s*'.format(_msg_type),
    entry=ColEntry
)


class SbtOutputEvent(OutputEvent, Location):
    col = maybe_field(ColEntry)
    file = field(FileEntry)

    @property
    def file_path(self):
        return self.file.file_path

    @property
    def coords(self):
        return self.file.line, self.col / _.col | 1

    @property
    def code(self):
        return self.entries.filter_type(CodeEntry)


class Parser(SimpleParser):

    @lazy
    def graph(self):
        g = DiGraph()
        g.add_edge('start', 'file', data=_file)
        g.add_edge('file', 'error', data=_code)
        g.add_edge('error', 'error', data=_code)
        g.add_edge('error', 'col', data=_col, weight=1)
        return g

    def event(self, entries: List[OutputEntry]):
        col = entries.find_type(ColEntry)
        return (
            entries.find_type(FileEntry) / (
                lambda a: SbtOutputEvent(head=List('error'), entries=entries,
                                         col=col, file=a)
            )
        )

__all__ = ('Parser',)



from typing import Callable

from amino import List, __, _

from lenses import lens

from ribosome.record import field, list_field, bool_field

from myo.record import Record
from myo.ui.tmux.session import Session, VimSession
from myo.ui.tmux.window import VimWindow, Window
from myo.ui.tmux.layout import VimLayout, Layout
from myo.ui.tmux.pane import VimPane
from myo.ui.tmux.util import Ident

_is_vim_session = lambda a: isinstance(a, VimSession)
_is_vim_window = lambda a: isinstance(a, VimWindow)
_is_vim_layout = lambda a: isinstance(a, VimLayout)
_is_vim_pane = lambda a: isinstance(a, VimPane)


class TmuxState(Record):
    initialized = bool_field()
    sessions = list_field()
    instance_id = field(str, initial='',
                        factory=lambda a: a if a else List.random_string(5))

    @property
    def _str_extra(self):
        return self.sessions

    def session_lens(self, pred: Callable[[Session], bool]):
        return self.attr_lens_pred(_.sessions, pred) / __.bind(self)

    @property
    def vim_session(self):
        return self.sessions.find(_is_vim_session).to_either('no vim session')

    @property
    def vim_session_lens(self):
        return (self.sessions.find_lens_pred(_is_vim_session) /
                lens(self).sessions.add_lens)

    def window_lens(self, pred: Callable[[Window], bool]):
        sub = __.attr_lens_pred(_.windows, pred)
        return (
            (self.attr_lens(_.sessions, sub) / __.bind(self))
            .to_either('no such window: {}'.format(pred))
        )

    def window_lens_ident(self, ident: Ident):
        return self.window_lens(__.has_ident(ident))

    @property
    def vim_window(self):
        return (
            (self.vim_session // __.windows.find(_is_vim_window))
            .to_either('no vim window')
        )

    @property
    def vim_window_lens(self):
        return self.window_lens(_is_vim_window)

    def layout_lens(self, pred: Callable[[Layout], bool]):
        lay_sub = __.root.layout_lens(pred).map(lens().root.add_lens)
        win_sub = __.attr_lens(_.windows, lay_sub)
        return self.attr_lens(_.sessions, win_sub) / __.bind(self)

    def layout_lens_ident(self, ident: Ident):
        return self.layout_lens(__.has_ident(ident))

    @property
    def vim_layout(self):
        return self.vim_window // __.root.layouts.find(_is_vim_layout)

    @property
    def vim_layout_lens(self):
        return self.layout_lens(_is_vim_layout)

    def pane_lens(self, pred: Callable[[Layout], bool]):
        lay_sub = __.root.pane_lens(pred).map(lens().root.add_lens)
        win_sub = __.attr_lens(_.windows, lay_sub)
        return self.attr_lens(_.sessions, win_sub) / __.bind(self)

    def pane_lens_ident(self, ident: Ident):
        return self.pane_lens(__.has_ident(ident))

    @property
    def vim_pane(self):
        return self.vim_layout / _.panes / _.head

    @property
    def all_windows(self):
        return self.sessions // _.windows

    def window(self, ident: Ident):
        return self.all_windows.find(__.has_ident(ident))

    @property
    def all_panes(self):
        return self.all_windows // _.root.all_panes

    def pane(self, ident):
        return self.all_panes.find(__.has_ident(ident))

    @property
    def possibly_open_panes(self):
        return (
            self.all_panes
            .filter(_.id.is_just)
            .filter(lambda a: not _is_vim_pane(a))
        )

__all__ = ('TmuxState',)
